from typing import Any
from collections import namedtuple
import datetime
from decimal import Decimal
import os
from pathlib import Path
import socket
import sys
from textwrap import dedent
from unittest import mock

from delta import configure_spark_with_delta_pip
import pytest

from pyspark.sql import SparkSession
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    MapType,
    NullType,
    ShortType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from koheesio.logger import LoggingFactory
from koheesio.spark.readers.dummy import DummyReader


def is_port_free(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("localhost", port))
            return True
        except socket.error:
            return False


@pytest.fixture(scope="session")
def warehouse_path(tmp_path_factory, random_uuid, logger):
    fldr = tmp_path_factory.mktemp("spark-warehouse" + random_uuid)
    logger.debug(f"Building test warehouse folder '{fldr}'")
    yield fldr.as_posix()


@pytest.fixture(scope="function")
def checkpoint_folder(tmp_path_factory, random_uuid, logger):
    fldr = tmp_path_factory.mktemp("checkpoint" + random_uuid)
    logger.debug(f"Building test checkpoint folder '{fldr}'")
    yield fldr.as_posix()


@pytest.fixture(scope="session")
def spark(warehouse_path, random_uuid):
    """Spark session fixture with Delta enabled."""
    builder = SparkSession.builder.appName("test_session" + random_uuid)

    if os.environ.get("SPARK_REMOTE") == "local":
        # SPARK_TESTING is set in environment variables
        # This triggers spark connect logic
        # ---->>>> For testing, we use 0 to use an ephemeral port to allow parallel testing.
        # --->>>>>> See also SPARK-42272.
        from pyspark.version import __version__ as spark_version

        builder = configure_spark_with_delta_pip(
            spark_session_builder=builder.remote("local"),
            extra_packages=[f"org.apache.spark:spark-connect_2.12:{spark_version}"],
        )
    else:
        builder = builder.master("local[*]")
        builder = configure_spark_with_delta_pip(spark_session_builder=builder)

    builder = (
        builder.config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
        .config("spark.sql.warehouse.dir", warehouse_path)
        .config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")
        .config("spark.sql.session.timeZone", "UTC")
        .config("spark.sql.execution.arrow.pyspark.enabled", "true")
        .config("spark.sql.execution.arrow.pyspark.fallback.enabled", "true")
    )

    spark_session = builder.getOrCreate()

    yield spark_session

    spark_session.stop()


@pytest.fixture(autouse=True, scope="session")
def set_env_vars():
    """
    Set environment variables for PYSPARK_PYTHON and PYSPARK_DRIVER_PYTHON.

    This function sets the environment variables PYSPARK_PYTHON and PYSPARK_DRIVER_PYTHON
    to the current Python executable. It is used to ensure that the correct Python version
    is used when running PySpark.
    """
    existing_pyspark_python = os.environ.get("PYSPARK_PYTHON")
    existing_pyspark_driver_python = os.environ.get("PYSPARK_DRIVER_PYTHON")

    os.environ["PYSPARK_PYTHON"] = sys.executable
    os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable

    yield

    if existing_pyspark_python is not None:
        os.environ["PYSPARK_PYTHON"] = existing_pyspark_python
    else:
        del os.environ["PYSPARK_PYTHON"]

    if existing_pyspark_driver_python is not None:
        os.environ["PYSPARK_DRIVER_PYTHON"] = existing_pyspark_driver_python
    else:
        del os.environ["PYSPARK_DRIVER_PYTHON"]


@pytest.fixture(scope="session", autouse=True)
def setup(spark, delta_file):
    db_name = "klettern"

    if not spark.catalog.databaseExists(db_name):
        spark.sql(f"CREATE DATABASE {db_name}")
        spark.sql(f"USE {db_name}")

    setup_test_data(spark=spark, delta_file=Path(delta_file))
    yield


@pytest.fixture(scope="class")
def dummy_df(spark):
    """
    | id |
    |----|
    | 1  |
    """
    yield DummyReader(range=1).read()


@pytest.fixture(scope="class")
def sample_df_to_partition(spark):
    """
    | partition | Value |
    |-----------|-------|
    | BE        | 12    |
    | FR        | 20    |
    """
    data = [["BE", 12], ["FR", 20]]
    schema = ["partition", "value"]
    return spark.createDataFrame(data, schema)


@pytest.fixture
def streaming_dummy_df(spark, delta_file):
    setup_test_data(spark=spark, delta_file=Path(delta_file))
    yield spark.readStream.table("delta_test_table")


@pytest.fixture(scope="class")
def sample_df_with_strings(spark):
    """
    df:
    | id | string |
    |----|--------|
    | 1  | hello  |
    | 2  | world  |
    | 3  |        |
    """
    data = [[1, "hello"], [2, "world"], [3, ""]]
    schema = ["id", "string"]
    return spark.createDataFrame(data, schema)


@pytest.fixture(scope="class")
def sample_df_with_timestamp(spark):
    """
    df:
    | id | a_date              | a_timestamp         |
    |----|---------------------|---------------------|
    | 1  | 1970-04-20 12:33:09 | 2000-07-01 01:01:00 |
    | 2  | 1980-05-21 13:34:08 | 2010-08-02 02:02:00 |
    | 3  | 1990-06-22 14:35:07 | 2020-09-03 03:03:00 |

    Schema:
    - id: bigint (nullable = true)
    - a_date: timestamp (nullable = true)
    - a_timestamp: timestamp (nullable = true)
    """
    data = [
        (1, datetime.datetime(1970, 4, 20, 12, 33, 9), datetime.datetime(2000, 7, 1, 1, 1)),
        (2, datetime.datetime(1980, 5, 21, 13, 34, 8), datetime.datetime(2010, 8, 2, 2, 2)),
        (3, datetime.datetime(1990, 6, 22, 14, 35, 7), datetime.datetime(2020, 9, 3, 3, 3)),
    ]
    schema = ["id", "a_date", "a_timestamp"]
    return spark.createDataFrame(data, schema)


@pytest.fixture(scope="class")
def sample_df_with_string_timestamp(spark):
    """
    | id | a_string_timestamp |
    |----|--------------------|
    | 1  | 202301010420       |
    | 2  | 202302020314       |
    """
    data = [(1, "202301010420"), (2, "202302020314")]
    schema = ["id", "a_string_timestamp"]
    return spark.createDataFrame(data, schema)


def setup_test_data(spark, delta_file):
    """
    Sets up test data for the Spark session. Reads a Delta file, creates a temporary view,
    and populates a Delta table with the view's data.
    """
    delta_file = delta_file.absolute().as_posix()
    spark.read.format("delta").load(delta_file).limit(10).createOrReplaceTempView("delta_test_view")
    spark.sql(
        dedent(
            """
            CREATE TABLE IF NOT EXISTS delta_test_table
            USING DELTA
            TBLPROPERTIES ("delta.enableChangeDataFeed" = "true")
            AS SELECT v.* FROM delta_test_view v
            """
        )
    )


SparkContextData = namedtuple("SparkContextData", ["spark", "options_dict"])
"""A named tuple containing the Spark session and the options dictionary used to create the DataFrame"""


@pytest.fixture(scope="class")
def dummy_spark(spark, sample_df_with_strings) -> SparkContextData:
    """SparkSession fixture that makes any call to SparkSession.read.load() return a DataFrame with strings.

    Because of the use of `type(spark.read)`, this fixture automatically alters its behavior for either a remote or
    regular Spark session.

    Example
    -------
    ```python
    def test_dummy_spark(dummy_spark, sample_df_with_strings):
        df = dummy_spark.read.load()
        assert df.count() == sample_df_with_strings.count()
    ```

    Returns
    -------
    SparkContextData
        A named tuple containing the Spark session and the options dictionary used to create the DataFrame
    """
    _options_dict = {}

    def mock_options(*args, **kwargs):
        _options_dict.update(kwargs)
        return spark.read

    spark_reader = type(spark.read)
    with mock.patch.object(spark_reader, "options", side_effect=mock_options):
        with mock.patch.object(spark_reader, "load", return_value=sample_df_with_strings):
            yield SparkContextData(spark, _options_dict)


@pytest.fixture(scope="class")
def mock_df(spark) -> mock.Mock:
    """Fixture to mock a DataFrame's methods."""
    # create a local DataFrame so we can get the spec of the DataFrame
    df = spark.range(1)

    # mock the df.write method
    mock_df_write = mock.create_autospec(type(df.write))

    # mock the save method
    mock_df_write.save = mock.Mock(return_value=None)

    # mock the format, option(s), and mode methods
    mock_df_write.format.return_value = mock_df_write
    mock_df_write.options.return_value = mock_df_write
    mock_df_write.option.return_value = mock_df_write
    mock_df_write.mode.return_value = mock_df_write

    # now create a mock DataFrame with the mocked write method
    mock_df = mock.create_autospec(type(df), instance=True)
    mock_df.write = mock_df_write
    yield mock_df


def await_job_completion(spark, timeout=300, query_id=None):
    """
    Waits for a Spark streaming job to complete.

    This function checks the status of a Spark streaming job and waits until it is completed or a timeout is reached.
    If a query_id is provided, it waits for the specific streaming job with that id. Otherwise, it waits for any active
    streaming job.
    """
    logger = LoggingFactory.get_logger(name="await_job_completion", inherit_from_koheesio=True)

    start_time = datetime.datetime.now()
    spark = spark.getActiveSession()
    logger.info("Waiting for streaming job to complete")
    if query_id is not None:
        stream = spark.streams.get(query_id)
        while stream.isActive and (datetime.datetime.now() - start_time).seconds < timeout:
            spark.streams.awaitAnyTermination(20)
    else:
        while len(spark.streams.active) > 0 and (datetime.datetime.now() - start_time).seconds < timeout:
            spark.streams.awaitAnyTermination(20)
    spark.streams.resetTerminated()
    logger.info("Streaming job completed")


@pytest.fixture
def df_with_all_types(spark):
    """Create a DataFrame with all supported Spark datatypes

    This DataFrame has 1 row and 14 columns, each column containing a single value of the given datatype.
    """
    data = dict(
        BYTE=(1, "byte", ByteType()),
        SHORT=(1, "short", ShortType()),
        INTEGER=(1, "integer", IntegerType()),
        LONG=(1, "long", LongType()),
        FLOAT=(1.0, "float", FloatType()),
        DOUBLE=(1.0, "double", DoubleType()),
        DECIMAL=(Decimal(1.0), "decimal", DecimalType()),
        STRING=("a", "string", StringType()),
        BINARY=(b"a", "binary", BinaryType()),
        BOOLEAN=(True, "boolean", BooleanType()),
        # '2023-01-01T00:01:01'
        TIMESTAMP=(datetime.datetime.utcfromtimestamp(1672531261), "timestamp", TimestampType()),
        DATE=(datetime.date(2023, 1, 1), "date", DateType()),
        ARRAY=(["a"], "array", ArrayType(StringType())),
        MAP=({"a": "b"}, "map", MapType(StringType(), StringType())),
        VOID=(None, "void", NullType()),
    )
    return spark.createDataFrame(
        data=[[v[0] for v in data.values()]],
        schema=StructType([StructField(name=v[1], dataType=v[2]) for v in data.values()]),
    )


class ScopeSecrets:
    class SecretMeta:
        def __init__(self, key: str):
            self.key = key

    def __init__(self, secrets: dict):
        self.secrets = secrets

    def get(self, scope: str, key: str) -> Any:
        return self.secrets.get(key)

    def list(self, scope: str):
        keys = [ScopeSecrets.SecretMeta(key=key) for key in self.secrets.keys()]

        return keys



"""
Adds a number of Transformations that are intended to be used with StringType column input.
Some will work with other types however, but will output StringType or an array of StringType.

These Transformations take full advantage of Koheesio's ColumnsTransformationWithTarget class, allowing a user to
apply column transformations to multiple columns at once. See the class docstrings for more information.

The following Transformations are included:

[change_case](change_case.md):

- `Lower`
    Converts a string column to lower case.
- `Upper`
    Converts a string column to upper case.
- `TitleCase` or `InitCap`
    Converts a string column to title case, where each word starts with a capital letter.

[concat](concat.md):

- `Concat`
    Concatenates multiple input columns together into a single column, optionally using the given separator.

[pad](pad.md):

- `Pad`
    Pads the values of `source_column` with the `character` up until it reaches `length` of characters
- `LPad`
    Pad with a character on the left side of the string.
- `RPad`
    Pad with a character on the right side of the string.

[regexp](regexp.md):

- `RegexpExtract`
    Extract a specific group matched by a Java regexp from the specified string column.
- `RegexpReplace`
    Searches for the given regexp and replaces all instances with what is in 'replacement'.

[replace](replace.md):

- `Replace`
    Replace all instances of a string in a column with another string.

[split](split.md):

- `SplitAll`
    Splits the contents of a column on basis of a split_pattern.
- `SplitAtFirstMatch`
    Like SplitAll, but only splits the string once. You can specify whether you want the first or second part.

[substring](substring.md):

- `Substring`
    Extracts a substring from a string column starting at the given position.

[trim](trim.md):

- `Trim`
    Trim whitespace from the beginning and/or end of a string.
- `LTrim`
    Trim whitespace from the beginning of a string.
- `RTrim`
    Trim whitespace from the end of a string.

"""



"""Lookup transformation for joining two dataframes together

Classes
-------
JoinMapping
TargetColumn
JoinType
JoinHint
DataframeLookup
"""

from typing import List, Optional, Union
from enum import Enum

from pyspark.sql import Column
from pyspark.sql import functions as f

from koheesio.models import BaseModel, Field, field_validator
from koheesio.spark import DataFrame
from koheesio.spark.transformations import Transformation

__all__ = [
    "JoinMapping",
    "TargetColumn",
    "JoinType",
    "JoinHint",
    "DataframeLookup",
]


class JoinMapping(BaseModel):
    """Mapping for joining two dataframes together"""

    source_column: str
    other_column: str

    @property
    def column(self) -> Column:
        """Get the join mapping as a pyspark.sql.Column object"""
        return f.col(self.other_column).alias(self.source_column)


class TargetColumn(BaseModel):
    """Target column for the joined dataframe"""

    target_column: str
    target_column_alias: str

    @property
    def column(self) -> Column:
        """Get the target column as a pyspark.sql.Column object"""
        return f.col(self.target_column).alias(self.target_column_alias)


class JoinType(str, Enum):
    """Supported join types"""

    INNER = "inner"
    FULL = "full"
    LEFT = "left"
    RIGHT = "right"
    CROSS = "cross"
    SEMI = "semi"
    ANTI = "anti"


class JoinHint(str, Enum):
    """Supported join hints"""

    BROADCAST = "broadcast"
    MERGE = "merge"


class DataframeLookup(Transformation):
    """Lookup transformation for joining two dataframes together

    Parameters
    ----------
    df : DataFrame
        The left Spark DataFrame
    other : DataFrame
        The right Spark DataFrame
    on : List[JoinMapping] | JoinMapping
        List of join mappings. If only one mapping is passed, it can be passed as a single object.
    targets : List[TargetColumn] | TargetColumn
        List of target columns. If only one target is passed, it can be passed as a single object.
    how : JoinType
        What type of join to perform. Defaults to left. See JoinType for more information.
    hint : JoinHint
        What type of join hint to use. Defaults to None. See JoinHint for more information.

    Example
    -------
    ```python
    from pyspark.sql import SparkSession
    from koheesio.spark.transformations.lookup import (
        DataframeLookup,
        JoinMapping,
        TargetColumn,
        JoinType,
    )

    spark = SparkSession.builder.getOrCreate()

    # create the dataframes
    left_df = spark.createDataFrame([(1, "A"), (2, "B")], ["id", "value"])
    right_df = spark.createDataFrame(
        [(1, "A"), (3, "C")], ["id", "value"]
    )

    # perform the lookup
    lookup = DataframeLookup(
        df=left_df,
        other=right_df,
        on=JoinMapping(source_column="id", joined_column="id"),
        targets=TargetColumn(
            target_column="value", target_column_alias="right_value"
        ),
        how=JoinType.LEFT,
    )

    output_df = lookup.transform()
    ```

    __output_df__:

    | id|value|right_value|
    |---|----|------------|
    |  1|    A|          A|
    |  2|    B|       null|

    In this example, the `left_df` and `right_df` dataframes are joined together using the `id` column. The `value`
    column from the `right_df` is aliased as `right_value` in the output dataframe.
    """

    df: Optional[DataFrame] = Field(default=None, description="The left Spark DataFrame")
    other: Optional[DataFrame] = Field(default=None, description="The right Spark DataFrame")
    on: Union[List[JoinMapping], JoinMapping] = Field(
        default=...,
        alias="join_mapping",
        description="List of join mappings. If only one mapping is passed, it can be passed as a single object.",
    )
    targets: Union[List[TargetColumn], TargetColumn] = Field(
        default=...,
        alias="target_columns",
        description="List of target columns. If only one target is passed, it can be passed as a single object.",
    )
    how: Optional[JoinType] = Field(
        default=JoinType.LEFT, description="What type of join to perform. Defaults to left. " + str(JoinType.__doc__)
    )
    hint: Optional[JoinHint] = Field(
        default=None, description="What type of join hint to use. Defaults to None. " + str(JoinHint.__doc__)
    )

    @field_validator("on", "targets")
    def set_list(cls, value: Union[List[JoinMapping], JoinMapping, List[TargetColumn], TargetColumn]) -> List:
        """Ensure that we can pass either a single object, or a list of objects"""
        return [value] if not isinstance(value, list) else value

    class Output(Transformation.Output):
        """Output for the lookup transformation"""

        left_df: DataFrame = Field(default=..., description="The left Spark DataFrame")
        right_df: DataFrame = Field(default=..., description="The right Spark DataFrame")

    def get_right_df(self) -> DataFrame:
        """Get the right side dataframe"""
        return self.other

    def execute(self) -> Output:
        """Execute the lookup transformation"""
        # prepare the right dataframe
        prepared_right_df = self.get_right_df().select(
            *[join_mapping.column for join_mapping in self.on],  # type: ignore
            *[target.column for target in self.targets],  # type: ignore
        )
        if self.hint:
            prepared_right_df = prepared_right_df.hint(self.hint)

        # generate the output
        self.output.left_df = self.df
        self.output.right_df = prepared_right_df
        self.output.df = self.df.join(
            prepared_right_df,  # type: ignore
            on=[join_mapping.source_column for join_mapping in self.on],
            how=self.how,
        )



"""
File writers for different formats:
- CSV
- Parquet
- Avro
- JSON
- ORC
- Text

The FileWriter class is a configurable Writer that allows writing to different file formats providing any option needed.
CsvFileWriter, ParquetFileWriter, AvroFileWriter, JsonFileWriter, OrcFileWriter, and TextFileWriter are convenience
classes that just set the `format` field to the corresponding file format.

"""

from __future__ import annotations

from typing import Union
from enum import Enum
from pathlib import Path

from koheesio.models import ExtraParamsMixin, Field, field_validator
from koheesio.spark.writers import BatchOutputMode, Writer

__all__ = [
    "FileFormat",
    "FileWriter",
    "CsvFileWriter",
    "ParquetFileWriter",
    "AvroFileWriter",
    "JsonFileWriter",
    "OrcFileWriter",
    "TextFileWriter",
]


class FileFormat(str, Enum):
    """Supported file formats for the FileWriter class"""

    csv = "csv"
    parquet = "parquet"
    avro = "avro"
    json = "json"
    orc = "orc"
    text = "text"


class FileWriter(Writer, ExtraParamsMixin):
    """
    A configurable Writer that allows writing to different file formats providing any option needed.

    Extra parameters can be passed to the writer as keyword arguments.

    Examples
    --------
    ```python
    writer = FileWriter(
        df=df,
        path="path/to/file.csv",
        output_mode=BatchOutputMode.APPEND,
        format=FileFormat.parquet,
        compression="snappy",
    )
    ```
    """

    output_mode: BatchOutputMode = Field(default=BatchOutputMode.APPEND, description="The output mode to use")
    format: FileFormat = Field(..., description="The file format to use when writing the data.")
    path: Union[Path, str] = Field(default=..., description="The path to write the file to")

    @field_validator("path")
    def ensure_path_is_str(cls, v: Union[Path, str]) -> str:
        """Ensure that the path is a string as required by Spark."""
        if isinstance(v, Path):
            return str(v.absolute().as_posix())
        return v

    def execute(self) -> FileWriter.Output:
        writer = self.df.write

        if self.extra_params:
            self.log.info(f"Setting extra parameters for the writer: {self.extra_params}")
            writer = writer.options(**self.extra_params)

        writer.save(path=self.path, format=self.format, mode=self.output_mode)  # type: ignore

        self.output.df = self.df


class CsvFileWriter(FileWriter):
    """Writes a DataFrame to a CSV file.

    This class is a convenience class that sets the `format` field to `FileFormat.csv`.

    Extra parameters can be passed to the writer as keyword arguments.

    Examples
    --------
    ```python
    writer = CsvFileWriter(
        df=df,
        path="path/to/file.csv",
        output_mode=BatchOutputMode.APPEND,
        header=True,
    )
    ```
    """

    format: FileFormat = FileFormat.csv


class ParquetFileWriter(FileWriter):
    """Writes a DataFrame to a Parquet file.

    This class is a convenience class that sets the `format` field to `FileFormat.parquet`.

    Extra parameters can be passed to the writer as keyword arguments.

    Examples
    --------
    ```python
    writer = ParquetFileWriter(
        df=df,
        path="path/to/file.parquet",
        output_mode=BatchOutputMode.APPEND,
        compression="snappy",
    )
    ```
    """

    format: FileFormat = FileFormat.parquet


class AvroFileWriter(FileWriter):
    """Writes a DataFrame to an Avro file.

    This class is a convenience class that sets the `format` field to `FileFormat.avro`.

    Extra parameters can be passed to the writer as keyword arguments.

    Examples
    --------
    ```python
    writer = AvroFileWriter(
        df=df,
        path="path/to/file.avro",
        output_mode=BatchOutputMode.APPEND,
    )
    ```
    """

    format: FileFormat = FileFormat.avro


class JsonFileWriter(FileWriter):
    """Writes a DataFrame to a JSON file.

    This class is a convenience class that sets the `format` field to `FileFormat.json`.

    Extra parameters can be passed to the writer as keyword arguments.

    Examples
    --------
    ```python
    writer = JsonFileWriter(
        df=df,
        path="path/to/file.json",
        output_mode=BatchOutputMode.APPEND,
    )
    ```
    """

    format: FileFormat = FileFormat.json


class OrcFileWriter(FileWriter):
    """Writes a DataFrame to an ORC file.

    This class is a convenience class that sets the `format` field to `FileFormat.orc`.

    Extra parameters can be passed to the writer as keyword arguments.

    Examples
    --------
    ```python
    writer = OrcFileWriter(
        df=df, path="path/to/file.orc", output_mode=BatchOutputMode.APPEND
    )
    ```
    """

    format: FileFormat = FileFormat.orc


class TextFileWriter(FileWriter):
    """Writes a DataFrame to a text file.

    This class is a convenience class that sets the `format` field to `FileFormat.text`.

    Extra parameters can be passed to the writer as keyword arguments.

    Examples
    --------
    ```python
    writer = TextFileWriter(
        df=df, path="path/to/file.txt", output_mode=BatchOutputMode.APPEND
    )
    ```
    """

    format: FileFormat = FileFormat.text



"""
Spark step module
"""

from __future__ import annotations

from typing import Optional
from abc import ABC
import warnings

from pydantic import Field

from koheesio import Step, StepOutput
from koheesio.models import model_validator
from koheesio.spark.utils.common import (
    AnalysisException,
    Column,
    DataFrame,
    DataFrameReader,
    DataFrameWriter,
    DataStreamReader,
    DataStreamWriter,
    DataType,
    ParseException,
    SparkSession,
    StreamingQuery,
)

__all__ = [
    "SparkStep",
    "Column",
    "DataFrame",
    "ParseException",
    "SparkSession",
    "AnalysisException",
    "DataType",
    "DataFrameReader",
    "DataStreamReader",
    "DataFrameWriter",
    "DataStreamWriter",
    "StreamingQuery",
]


class SparkStep(Step, ABC):
    """Base class for a Spark step

    Extends the Step class with SparkSession support. The following:
    - Spark steps are expected to return a Spark DataFrame as output.
    - spark property is available to access the active SparkSession instance.
    - The SparkSession instance can be provided as an argument to the constructor through the `spark` parameter.
    """

    spark: Optional[SparkSession] = Field(
        default=None,
        description="The SparkSession instance. If not provided, the active SparkSession will be used.",
        validate_default=False,
    )

    class Output(StepOutput):
        """Output class for SparkStep"""

        df: Optional[DataFrame] = Field(default=None, description="The Spark DataFrame")

    @model_validator(mode="after")
    def _get_active_spark_session(self) -> SparkStep:
        """Return active SparkSession instance
        If a user provides a SparkSession instance, it will be returned. Otherwise, an active SparkSession will be
        attempted to be retrieved.
        """
        if self.spark is None:
            from koheesio.spark.utils.common import get_active_session

            self.spark = get_active_session()
        return self


def current_timestamp_utc(spark):
    warnings.warn(
        message=(
            "The current_timestamp_utc function has been moved to the koheesio.spark.functions module."
            "Import it from there instead. Current import path will be deprecated in the future."
        ),
        category=DeprecationWarning,
        stacklevel=2,
    )
    from koheesio.spark.functions import current_timestamp_utc as _current_timestamp_utc

    return _current_timestamp_utc(spark)



"""
Create Spark DataFrame directly from the data stored in a Python variable
"""

from typing import Any, Dict, Optional, Union
from enum import Enum
from functools import partial
from io import StringIO
import json

import pandas as pd

from pyspark.sql.types import StructType

from koheesio.models import ExtraParamsMixin, Field
from koheesio.spark import DataFrame
from koheesio.spark.readers import Reader


class DataFormat(Enum):
    """Data formats supported by the InMemoryDataReader"""

    JSON = "json"
    CSV = "csv"


class InMemoryDataReader(Reader, ExtraParamsMixin):
    """Directly read data from a Python variable and convert it to a Spark DataFrame.

    Read the data, that is stored in one of the supported formats (see `DataFormat`) directly from the variable and
    convert it to the Spark DataFrame. The use cases include converting JSON output of the API into the dataframe;
    reading the CSV data via the API (e.g. Box API).

    The advantage of using this reader is that it allows to read the data directly from the Python variable, without
    the need to store it on the disk. This can be useful when the data is small and does not need to be stored
    permanently.

    Parameters
    ----------
    data : Union[str, list, dict, bytes]
        Source data
    format : DataFormat
        File / data format
    schema_ : Optional[StructType], optional, default=None
        Schema that will be applied during the creation of Spark DataFrame
    params : Optional[Dict[str, Any]], optional, default=dict
        Set of extra parameters that should be passed to the appropriate reader (csv / json). Optionally, the user can
        pass the parameters that are specific to the reader (e.g. `multiLine` for JSON reader) as key-word arguments.
        These will be merged with the `params` parameter.

    Example
    -------
    ```python
    # Read CSV data from a string
    df1 = InMemoryDataReader(format=DataFormat.CSV, data='foo,bar\\nA,1\\nB,2')

    # Read JSON data from a string
    df2 = InMemoryDataReader(format=DataFormat.JSON, data='{"foo": A, "bar": 1}'
    df3 = InMemoryDataReader(format=DataFormat.JSON, data=['{"foo": "A", "bar": 1}', '{"foo": "B", "bar": 2}']
    ```
    """

    data: Union[str, list, dict, bytes] = Field(default=..., description="Source data")
    format: DataFormat = Field(default=..., description="File / data format")
    schema_: Optional[StructType] = Field(
        default=None,
        alias="schema",
        description="[Optional] Schema that will be applied during the creation of Spark DataFrame",
    )

    params: Dict[str, Any] = Field(
        default_factory=dict,
        description="[Optional] Set of extra parameters that should be passed to the appropriate reader (csv / json)",
    )

    def _csv(self) -> DataFrame:
        """Method for reading CSV data"""
        if isinstance(self.data, list):
            csv_data: str = "\n".join(self.data)
        else:
            csv_data: str = self.data  # type: ignore

        if "header" in self.params and self.params["header"] is True:
            self.params["header"] = 0

        pandas_df = pd.read_csv(StringIO(csv_data), **self.params)  # type: ignore
        df = self.spark.createDataFrame(pandas_df, schema=self.schema_)  # type: ignore

        return df

    def _json(self) -> DataFrame:
        """Method for reading JSON data"""
        if isinstance(self.data, str):
            json_data = [json.loads(self.data)]
        elif isinstance(self.data, list):
            if all(isinstance(x, str) for x in self.data):
                json_data = [json.loads(x) for x in self.data]
            else:
                json_data = self.data
        else:
            json_data = [self.data]

        # Use pyspark.pandas to read the JSON data from the string
        # noinspection PyUnboundLocalVariable
        pandas_df = pd.read_json(StringIO(json.dumps(json_data)), **self.params)  # type: ignore

        # Convert pyspark.pandas DataFrame to Spark DataFrame
        df = self.spark.createDataFrame(pandas_df, schema=self.schema_)  # type: ignore

        return df

    def execute(self) -> Reader.Output:
        """
        Execute method appropriate to the specific data format
        """
        if self.data is None:
            raise ValueError("Data is not provided")

        if isinstance(self.data, bytes):
            self.data = self.data.decode("utf-8")

        _func = getattr(InMemoryDataReader, f"_{self.format}")
        _df = partial(_func, self)()
        self.output.df = _df



"""
HANA reader.
"""

from typing import Any, Dict, Optional

from koheesio.models import Field
from koheesio.spark.readers.jdbc import JdbcReader


class HanaReader(JdbcReader):
    """
    Wrapper around JdbcReader for SAP HANA

    Notes
    -----
    * Refer to [JdbcReader](jdbc.md#koheesio.spark.readers.jdbc.JdbcReader) for the list of all available parameters.
    * Refer to SAP HANA Client Interface Programming Reference docs for the list of all available connection string
        parameters:
        https://help.sap.com/docs/SAP_HANA_CLIENT/f1b440ded6144a54ada97ff95dac7adf/109397c2206a4ab2a5386d494f4cf75e.html

    Example
    -------
    > Note: jars should be added to the Spark session manually. This class does not take care of that.

    This example depends on the SAP HANA `ngdbc` JAR. e.g. ngdbc-2.5.49.

    ```python
    from koheesio.spark.readers.hana import HanaReader

    jdbc_hana = HanaReader(
        url="jdbc:sap://<domain_or_ip>:<port>/?<options>",
        user="YOUR_USERNAME",
        password="***",
        dbtable="schema_name.table_name",
    )
    df = jdbc_hana.read()
    ```

    Parameters
    ----------
    url : str
        JDBC connection string. Refer to SAP HANA docs for the list of all available connection string parameters.
        Example: jdbc:sap://<server>:<port>[/?<options>]
    user : str
    password : SecretStr
    dbtable : str
        Database table name, also include schema name
    params : Optional[Dict[str, Any]], default={"fetchsize": 2000, "numPartitions": 10}
        Extra options/params to pass to the SAP HANA JDBC driver. Refer to SAP HANA docs for the list of all available
        connection string parameters. Also can be passed as arbitrary keyword arguments to the HanaReader class.
        Default values: {"fetchsize": 2000, "numPartitions": 10}
    query : Optional[str]
        Query
    format : str
        The type of format to load. Defaults to 'jdbc'. Should not be changed.
    driver : str
        Driver name. Be aware that the driver jar needs to be passed to the task. Should not be changed.
    """

    driver: str = Field(
        default="com.sap.db.jdbc.Driver",
        description="Make sure that the necessary JARs are available in the cluster: ngdbc-2-x.x.x.x",
    )
    params: Optional[Dict[str, Any]] = Field(
        default={"fetchsize": 2000, "numPartitions": 10},
        description="Extra options to pass to the SAP HANA JDBC driver",
        alias="options",
    )



"""
Extracts a substring from a string column starting at the given position.
"""

from pyspark.sql import Column
from pyspark.sql.functions import substring, when
from pyspark.sql.types import StringType

from koheesio.models import Field, PositiveInt, field_validator
from koheesio.spark.transformations import ColumnsTransformationWithTarget


class Substring(ColumnsTransformationWithTarget):
    """
    Extracts a substring from a string column starting at the given position.

    This is a wrapper around PySpark substring() function

    Notes
    -----
    - Numeric columns will be cast to string
    - start is 1-indexed, not 0-indexed!

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to substring. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    start : PositiveInt
        Positive int. Defines where to begin the substring from. The first character of the field has index 1!
    length : Optional[int], optional, default=-1
        Optional. If not provided, the substring will go until end of string.

    Example
    -------
    ### Extract a substring from a string column starting at the given position.

    __input_df:__

    | column    |
    |-----------|
    | skyscraper|

    ```python
    output_df = Substring(
        column="column",
        target_column="substring_column",
        start=3,  # 1-indexed! So this will start at the 3rd character
        length=4,
    ).transform(input_df)
    ```

    __output_df:__

    | column    | substring_column |
    |-----------|------------------|
    | skyscraper| yscr             |

    """

    start: PositiveInt = Field(default=..., description="The starting position")
    length: int = Field(
        default=-1,
        description="The target length for the string. use -1 to perform until end",
    )

    @field_validator("length")
    def _valid_length(cls, length_value: int) -> int:
        """Integer.maxint fix for Java.
        Python's sys.maxsize is larger which makes f.substring fail"""
        if length_value == -1:
            return 2147483647
        return length_value

    def func(self, column: Column) -> Column:
        return when(column.isNull(), None).otherwise(substring(column, self.start, self.length)).cast(StringType())
